/*
 * Copyright (c) 2006 Ed Schouten <ed@fxq.nl>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: audio.c 479 2006-08-01 07:06:42Z ed $
 */
/**
 * @file audio/audio.c
 */

#include "config.h"

#include <stdlib.h>
#include <string.h>

#include <audio/audio.h>

#include <audio/vorbis/driver.h>
#include <audio/mp3/driver.h>
#include <audio/sndfile/driver.h>

/*
 * Make sure you add the most strict matching drivers at the top. Raw or
 * headerless file formats should be listed at the bottom, as the
 * matching code just runs through this list.
 */
/**
 * @brief List of audio drivers.
 */
static struct audio_driver drivers[] = {
#ifdef BUILD_VORBIS
	{ "Ogg Vorbis", vorbis_open, vorbis_close, vorbis_read, vorbis_seek },
#endif /* !BUILD_VORBIS */
#ifdef BUILD_MP3
	{ "MP3",	mp3_open, mp3_close, mp3_read, NULL },
#endif /* !BUILD_MP3 */
	/*
	 * Keep this entry at the bottom - it does evil stuff with raw
	 * file descriptors. It could also catch some raw formats.
	 */
#ifdef BUILD_SNDFILE
	{ "libsndfile",	sndfile_open, sndfile_close, sndfile_read, NULL },
#endif /* !BUILD_SNDFILE */
};
/**
 * @brief Amount of audio drivers.
 */
#define NUM_DRIVERS (sizeof(drivers) / sizeof(struct audio_driver))

void
audio_dump_formats(void)
{
	int i;

	for (i = 0; i < NUM_DRIVERS; i++)
		fprintf(stderr, " - %s\n", drivers[i].type);
}

struct audio_file *
audio_open(const char *filename)
{
	struct audio_file *out = NULL;
	int i;

	out = malloc(sizeof(struct audio_file));

	if (out == NULL)
		return (NULL);
	
	memset(out, 0x00000000, sizeof(struct audio_file));
	
	out->fp = fopen(filename, "rb");
	if (out->fp == NULL)
		goto bad;

	for (i = 0; i < NUM_DRIVERS; i++) {
		rewind(out->fp);

		if (drivers[i].open(out) == 0) {
			/* Assign the driver to the file */
			out->drv = &drivers[i];
			break;
		}
	}

	if (out->drv == NULL) {
		/* No matched driver */
		fclose(out->fp);
		goto bad;
	}

	return (out);

bad:
	free(out);
	return (NULL);
}

void
audio_close(struct audio_file *fd)
{
	fd->drv->close(fd);
	if (fd->fp != NULL)
		fclose(fd->fp);

	free(fd);
}

void
audio_seek(struct audio_file *fd, double len)
{
	if ((len != 0) && (fd->drv->seek != NULL))
		fd->drv->seek(fd, len);
}
