/*
 * Copyright (c) 2001--2004 Bertrand Petit
 *
 * Copyright (c) 2006 Ed Schouten <ed@fxq.nl>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 
 * 3. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: mad.c 482 2006-08-01 08:17:58Z ed $
 */
/**
 * @file audio/mp3/mad.c
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>

#include <mad.h>

#include <audio/mp3/driver.h>
#include <audio/output/output.h>

/* ntohl() */
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif /* HAVE_ARPA_INET_H */
#ifdef HAVE_WINSOCK2_H
#include <winsock2.h>
#endif /* HAVE_WINSOCK2_H */

/*
 * For some reason, libmad isn't documented at all. There's only an
 * example application called madlld. This code is loosely based on
 * madlld.c. The key differences are:
 *
 * - Audio is written to a buffer, instead of a file
 * - The code is written to return PCM of a requested size, instead of
 *   handling it frame by frame.
 * - It generates 16 bits signed little endian PCM, instead of big
 *   endian
 */

/**
 * @brief Private MP3 data stored in the audio file structure.
 */
struct mp3_drv_data {
	/**
	 * @brief Stream data.
	 */
	struct mad_stream	mstream;
	/**
	 * @brief Frame data.
	 */
	struct mad_frame	mframe;
	/**
	 * @brief Synth data.
	 */
	struct mad_synth	msynth;
	/**
	 * @brief Progress timer.
	 */
	mad_timer_t		mtimer;
	/**
	 * @brief Sample offset.
	 */
	int			cursample;

	/**
	 * @brief Read buffer.
	 */
	unsigned char buf_input[65536];
};

/**
 * @brief Parse the ID3 tag from the MP3 file.
 */
static void
mp3_parse_tag(struct audio_file *fd)
{
	/*
	 * This MP3 tag parsing routine is very simple (and incomplete).
	 * We should really make use of some specialized library to
	 * parse tags.
	 */
	
	unsigned long len = 0, rlen;
	char tag_type[4];
	char *value;
	
	/* Skip the version */
	if (fseek(fd->fp, 4, SEEK_SET) == -1)
		return;
	if (fread(&len, sizeof(len), 1, fd->fp) != 1)
		return;
	if (fseek(fd->fp, 2 + ntohl(len), SEEK_CUR) == -1)
		return;

	for (;;) {
		/* Read the tag type */
		if (fread(tag_type, sizeof(tag_type), 1, fd->fp) != 1)
			return;
		if (tag_type[0] == '\0')
			return;

		/* Length and flags */
		if (fread(&len, sizeof(len), 1, fd->fp) != 1)
			return;
		/* XXX */
		fseek(fd->fp, 3, SEEK_CUR);

		/* Read the value */
		if ((rlen = ntohl(len)) <= 1)
			continue;
		if ((value = malloc(rlen)) == NULL)
			return;
		if (fread(value, rlen - 1, 1, fd->fp) != 1) {
			free(value);
			return;
		}
		value[rlen - 1] = '\0';

		if (strncmp(tag_type, "TPE1", sizeof(tag_type)) == 0)
			fd->tag.artist = value;
		else if (strncmp(tag_type, "TIT2", sizeof(tag_type)) == 0)
			fd->tag.title = value;
		else if (strncmp(tag_type, "TALB", sizeof(tag_type)) == 0)
			fd->tag.album = value;
		else
			free(value);
	}
}

int
mp3_open(struct audio_file *fd)
{
	struct mp3_drv_data *data;
	char magic[3];

	/* Primitive magic check */
	if (fread(magic, 1, sizeof(magic), fd->fp) != sizeof(magic))
		return (-1);
	if (strncmp(magic, "ID3", sizeof(magic)) != 0)
		return (-1);

	if ((data = malloc(sizeof(struct mp3_drv_data))) == NULL)
		return (-1);

	data->cursample = 0;

	mad_stream_init(&data->mstream);
	mad_frame_init(&data->mframe);
	mad_synth_init(&data->msynth);
	mad_timer_reset(&data->mtimer);

	fd->drv_data = (void *)data;

	mp3_parse_tag(fd);
	
	return (0);
}

void
mp3_close(struct audio_file *fd)
{
	struct mp3_drv_data *data = fd->drv_data;
	
	mad_frame_finish(&data->mframe);
	mad_stream_finish(&data->mstream);
	mad_synth_finish(&data->msynth);

	free(data);
}

/**
 * @brief Refill the databuffer when it's drained, copying the last
 *        partial frame to the beginning.
 */
static size_t
mp3_fill_buffer(struct audio_file *fd)
{
	struct mp3_drv_data *data = fd->drv_data;
	size_t offset, filledlen, readlen;

	if (data->mstream.next_frame == NULL) {
		/* Place contents at the beginning */
		offset = 0;
	} else {
		/*
		 * We still need to save the fragmented frame. Copy it to the
		 * beginning and load as much as possible.
		 */
		offset = data->mstream.bufend - data->mstream.next_frame;
		memmove(data->buf_input, data->mstream.next_frame, offset);
	}

	readlen = sizeof(data->buf_input) - offset;

	filledlen = fread(data->buf_input + offset, 1, readlen, fd->fp);

	if (filledlen <= 0)
		return (0);

	if (filledlen < readlen) {
		/*
		 * Place some null bytes at the end
		 * because we may overrun it
		 */
		memset(data->buf_input + offset + filledlen, 0x00000000,
		    MAD_BUFFER_GUARD);
		filledlen += MAD_BUFFER_GUARD;
	}

	/* Return the length of the entire buffer */
	return (offset + filledlen);
}

/**
 * @brief Try to decode a frame, refilling the buffer if needed.
 */
static int
mp3_read_frame(struct audio_file *fd)
{
	struct mp3_drv_data *data = fd->drv_data;
	size_t buflen;

	/*  Get the next frame */
	for (;;) {
		/* We've run out of data. Read from file */
		if ((data->mstream.buffer == NULL) ||
		    (data->mstream.error == MAD_ERROR_BUFLEN)) {
			if ((buflen = mp3_fill_buffer(fd)) == 0)
				/* Read error */
				return (1);

			/* Load it back in */
			mad_stream_buffer(&data->mstream, data->buf_input,
			    buflen);
			data->mstream.error = 0;
		}

		if (mad_frame_decode(&data->mframe, &data->mstream) == 0)
			/* We're done */
			break;
	
		if (!MAD_RECOVERABLE(data->mstream.error) &&
		    (data->mstream.error != MAD_ERROR_BUFLEN))
			/* Unrecoverable error - bail out */
			return (1);
	}

	/* Update the audio timer */
	mad_timer_add(&data->mtimer, data->mframe.header.duration);
	fd->time_cur = data->mtimer.seconds;

	/* XXX: Quite crude - no way to determine the actual time */
	if (fd->time_len < fd->time_cur)
		fd->time_len = fd->time_cur;

	return (0);
}

/**
 * @brief Convert a fixed point sample to a short.
 */
static signed short
mp3_fixed_to_short(mad_fixed_t fixed)
{
	signed short sample;

	if (fixed >= MAD_F_ONE)
		sample = SHRT_MAX;
	else if (fixed <= -MAD_F_ONE)
		sample = -SHRT_MAX;
	else
		sample = fixed >> (MAD_F_FRACBITS - 15);

	return (sample);
}

/**
 * @brief Fetch a sample converted to short from the synth.
 */
#define mp3_calc_sample(chan,idx) \
    mp3_fixed_to_short(data->msynth.pcm.samples[chan][idx])

size_t
mp3_read(struct audio_file *fd, void *buf)
{
	struct mp3_drv_data *data = fd->drv_data;
	size_t written = 0;
	unsigned char *out = buf;

	do {
		/* Get a new frame when we haven't go one */
		if (data->cursample == 0) {
			if (mp3_read_frame(fd) != 0)
				goto done;

			/* We can now set the sample rate */
			fd->srate = data->mframe.header.samplerate;

			mad_synth_frame(&data->msynth, &data->mframe);
		}

		while ((data->cursample < data->msynth.pcm.length) &&
		    (written < AUDIO_OUTPUT_BUFLEN)) {
			signed short sample;

			/* Left channel */
			sample = mp3_calc_sample(0, data->cursample);
			out[written++] = sample & 0xff;
			out[written++] = sample >> 8;

			/* Right channel (if available) */
			if (MAD_NCHANNELS(&data->mframe.header) == 2)
				sample = mp3_calc_sample(1, data->cursample);
			out[written++] = sample & 0xff;
			out[written++] = sample >> 8;

			data->cursample++;
		}

		/* Move on to the next frame */
		if (data->cursample == data->msynth.pcm.length)
			data->cursample = 0;

	} while (written < AUDIO_OUTPUT_BUFLEN);

done:
	return (written);
}
