/*
 * Copyright (c) 2006 Ed Schouten <ed@fxq.nl>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: configfile.c 479 2006-08-01 07:06:42Z ed $
 */
/**
 * @file configfile/configfile.c
 */

#include <sys/param.h>
#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <curses.h>

#include <configfile/configfile.h>
#include <missing/missing.h>

#if 0
static int config_altered = 0;
#endif

/**
 * @brief Structure containing a single configuration entry of Herrie
 */
struct config_entry {
	/**
	 * @brief Name it can be located at
	 */
	const char	*name;
	/**
	 * @brief Default value
	 */
	const char	*defval;
	/**
	 * @brief Function to validate when set
	 */
	int		(*validator)(char *val);
	/**
	 * @brief Current value if not equal to the default
	 */
	char		*curval;
};

/* Config switch validators */
static int valid_bool(char *val);
#if 0
static int valid_md5(char *val);
#endif
static int valid_color(char *val);

/*
 * List of available configuration switches. Keep this list sorted at
 * all time; the config lookups have been optimised to expect an
 * alphabetical lookup. Looking up switches starting with 'z' will take
 * longer than 'a'.
 */
/**
 * @brief List of configuration switches.
 */
static struct config_entry configlist[] = {
	{ "audio.output.alsa.device",	"default",	NULL, 		NULL },
	{ "audio.output.oss.device",	"/dev/dsp",	NULL,		NULL },
	{ "gui.browser.defaultpath",	"",		NULL,		NULL },
	{ "gui.color.bar.bg",		"blue",		valid_color,	NULL },
	{ "gui.color.bar.fg",		"white",	valid_color,	NULL },
	{ "gui.color.block.bg",		"black",	valid_color,	NULL },
	{ "gui.color.block.fg",		"white",	valid_color,	NULL },
	{ "gui.color.deselect.bg",	"white",	valid_color,	NULL },
	{ "gui.color.deselect.fg",	"black",	valid_color,	NULL },
	{ "gui.color.select.bg",	"cyan",		valid_color,	NULL },
	{ "gui.color.select.fg",	"black",	valid_color,	NULL },
#if 0
	{ "scrobbler.password",		"",		valid_md5,	NULL },
	{ "scrobbler.username",		"",		NULL,		NULL },
#endif
	{ "vfs.dir.hide_dotfiles",	"yes",		valid_bool,	NULL },
	{ "vfs.lockup.chroot",		"",		NULL,		NULL },
	{ "vfs.lockup.user",		"",		NULL,		NULL },
};
/**
 * @brief The amount of configuration switches available.
 */
#define CONFIGENTRIES (sizeof(configlist) / sizeof(struct config_entry)) 

/**
 * @brief Search for an entry in the configlist by name.
 */
static struct config_entry *
config_search(char *opt)
{
	int i, c;

	for (i = 0; i < CONFIGENTRIES; i++) {
		c = strcmp(opt, configlist[i].name);

		if (c == 0)
			/* Found it */
			return (&configlist[i]);
		else if (c < 0)
			/* We're already too far */
			break;
	}

	/* Not found */
	return (NULL);
}

int
config_setopt(char *opt, char *val)
{
	struct config_entry *ent;
	char *newval = NULL;

	if ((ent = config_search(opt)) == NULL)
		return (-1);
	
	if (strcmp(val, ent->defval) != 0) {
		/* Just unset the value when it's the default */

		if (ent->validator != NULL) {
			/* Do not set invalid values */
			if (ent->validator(val) != 0)
				return (-1);
		}
		if ((newval = strdup(val)) == NULL)
			return (-1);
	}

	if (ent->curval != NULL)
		/* Free the current contents */
		free (ent->curval);

	ent->curval = newval;

#if 0
	config_altered = 1;
#endif
	return (0);
}

const char *
config_getopt(char *opt)
{
	struct config_entry *ent;
	
	ent = config_search(opt);
	assert(ent != NULL);
	
	/* Return the default if it is unset */
	return (ent->curval ? ent->curval : ent->defval);
}

void
config_load(char *file)
{
	FILE *cfile;
	char cfile_name[PATH_MAX], *homedir, *cline_opt, *cline_val;
	size_t clen;

	if (file == NULL) {
		/* No filename specified - search in homedir */
		if ((homedir = getenv("HOME")) == NULL)
			return;
		
		snprintf(cfile_name, PATH_MAX, "%s/.herrie/config", homedir);
		file = cfile_name;
	}

	if ((cfile = fopen(file, "rt")) == NULL)
		return;

	while ((cline_val = fgetln(cfile, &clen)) != NULL) {
		/* Null terminate the string first */
		cline_val[clen - 1] = '\0';

		/* Split at the : and set the option */
		cline_opt = strsep(&cline_val, ":");
		if (cline_val != NULL)
			config_setopt(cline_opt, cline_val);
	}

	fclose(cfile);

#if 0
	config_altered = 0;
#endif
}

#if 0
int
config_save(void)
{
	int i;
	FILE *cfile;

	if (!config_altered)
		return (0);

	if ((cfile = fopen(CONFIG_FILE, "wt")) == NULL)
		return (-1);
	
	for (i = 0; i < CONFIGENTRIES; i++) {
		if (configlist[i].curval != NULL) {
			fprintf(cfile, "%s:%s\n",
			    configlist[i].name,
			    configlist[i].curval);
		}
	}
	
	fclose(cfile);
	
	return (0);
}
#endif

/**
 * @brief Convert a "yes"/"no" string to a boolean value.
 */
static int
string_to_bool(const char *val)
{
	if (strcmp(val, "yes") == 0)
		return (1);
	else if (strcmp(val, "no") == 0)
		return (0);
	else
		return (-1);
}

/**
 * @brief Convert a color name to a Curses color code.
 */
static int
string_to_curses_color(const char *val)
{
	if (strcmp(val, "black") == 0)
		return (COLOR_BLACK);
	else if (strcmp(val, "red") == 0)
		return (COLOR_RED);
	else if (strcmp(val, "green") == 0)
		return (COLOR_GREEN);
	else if (strcmp(val, "yellow") == 0)
		return (COLOR_YELLOW);
	else if (strcmp(val, "blue") == 0)
		return (COLOR_BLUE);
	else if (strcmp(val, "magenta") == 0)
		return (COLOR_MAGENTA);
	else if (strcmp(val, "cyan") == 0)
		return (COLOR_CYAN);
	else if (strcmp(val, "white") == 0)
		return (COLOR_WHITE);
	
	return (-1);
}

/*
 * Configuration value validation
 */

/**
 * @brief Determine if a boolean string is valid
 */
static int valid_bool(char *val)
{
	return (string_to_bool(val) == -1);
}

#if 0
static int
valid_md5(char *val)
{
	int i;

	if (val[0] == '\0')
		return (0);
	
	for (i = 0; i < 32; i++) {
		/* Silently convert hash to lowercase */
		val[i] = tolower(val[i]);

		if (!isdigit(val[i]) &&
		    !(val[i] >= 'a' && val[i] <= 'f'))
			/* Invalid character found */
			return (-1);
	}

	if (val[i] != '\0')
		/* String too long */
		return (-1);

	return (0);
}
#endif

/**
 * @brief Determine if a color string is valid
 */
static int
valid_color(char *val)
{
	return ((string_to_curses_color(val) == -1) ? -1 : 0);
}

/*
 * Configuration value translations
 */

int
config_getopt_bool(char *val)
{
	int bval;
	
	bval = string_to_bool(config_getopt(val));
	assert(bval != -1);
	return (bval);
}

int
config_getopt_color(char *val)
{
	int col;
	
	col = string_to_curses_color(config_getopt(val));
	assert(col >= 0);
	return (col);
}
