/*
 * Copyright (c) 2006 Ed Schouten <ed@fxq.nl>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: playq.c 481 2006-08-01 07:27:53Z ed $
 */
/**
 * @file playq/playq.c
 */

#include "config.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>

#include <missing/missing.h>
#include <missing/semaphore.h>
#include <playq/playq.h>
#include <audio/audio.h>
#include <audio/output/output.h>
#include <gui/gui.h>
#include <missing/queue.h>

int playq_len;
struct playq_head playq_list =
    TAILQ_HEAD_INITIALIZER(playq_list);

pthread_mutex_t			playq_lock;
/**
 * @brief Semaphore holding the amount which the playback thread may
 *        'shift' songs off the playlist.
 */
static sem_t			*playq_available;

/* Playlist modifier commands */
static volatile int		do_pause;
static volatile int		do_skip;
static volatile double		do_seek;

/**
 * @brief Deallocate a playlist entry.
 */
static void
playq_entry_free(struct playq_entry *ent)
{
	free(ent->filename);
	free(ent);
}

/**
 * @brief Allocate a playlist entry.
 */
static struct playq_entry *
playq_entry_alloc(char *filename, char *name)
{
	struct playq_entry *ent;

	ent = malloc(sizeof(struct playq_entry));

	if (ent == NULL)
		return (NULL);
	
	if ((ent->filename = strdup(filename)) == NULL) {
		free(ent);
		return (NULL);
	}

	/* Evil pointer voodoo - copy offsets from original name */
	ent->name = ent->filename + (name - filename);

	return (ent);
}

/**
 * @brief Reference to playback thread.
 */
static pthread_t		playq_runner;

/**
 * @brief Infinitely play music in the playlist, honouring the do_
 *        flags.
 */
static void *
playq_runner_thread(void *unused)
{
	struct playq_entry	*new;
	struct audio_file	*cur;
#ifdef SIGWINCH
	sigset_t sset;

	/*
	 * Make sure we don't catch SIGWINCH'es as they are meant for
	 * the inputloop
	 */
	sigaddset(&sset, SIGWINCH);
	pthread_sigmask(SIG_BLOCK, &sset, NULL);
#endif /* !SIGWINCH */

	for (;;) {
		/* Fetch the next song */
		sem_wait(playq_available);
		PLAYQ_LOCK;
		gui_playq_notify_pre_removal(0);
		new = TAILQ_FIRST(&playq_list);
		TAILQ_REMOVE(&playq_list, new, next);
		playq_len--;
		gui_playq_notify_done();
		PLAYQ_UNLOCK;

		assert(new != NULL);

		cur = audio_open(new->filename);
		playq_entry_free(new);
		
		/* Skip broken songs */
		if (cur == NULL)
			continue;

		do_pause = do_skip = do_seek = 0;

		do  {
			while (do_pause)
				sleep(1);

			if (do_seek != 0) {
				audio_seek(cur, do_seek);
				do_seek = 0;
			}

			gui_statbar_update(cur);
		} while (!do_skip && (audio_output_play(cur) > 0));

		gui_statbar_update(NULL);
		audio_close(cur);
	}
}

int
playq_init(void)
{
	/* We need this for randomizing the playlist */
	srandomdev();

	if (audio_output_open() != 0)
		return (-1);

	pthread_mutex_init(&playq_lock, NULL);

	playq_available = sem_wrap_init("playq", 0);
	if (playq_available == NULL)
		return (-1);

	pthread_create(&playq_runner, NULL, playq_runner_thread, NULL);

	return (0);
}

void
playq_destroy(void)
{
	/* XXX: Clean up the queue */
	pthread_cancel(playq_runner);
	pthread_mutex_destroy(&playq_lock);

	sem_wrap_destroy(playq_available);

	audio_output_close();
}

/*
 * Public queue functions: playq_song_*
 */

/**
 * @brief Recursively add a VFS entity to the head of the playlist.
 */
static void
do_playq_song_add_head(struct vfs_ent *ve)
{
	struct playq_entry *ent;
	struct vfs_ent *cve;

	if (vfs_playable(ve)) {
		ent = playq_entry_alloc(ve->filename, ve->name);

		if (ent != NULL) {
			TAILQ_INSERT_HEAD(&playq_list, ent, next);
			playq_len++;
			gui_playq_notify_post_insertion(0);
			sem_post(playq_available);
		}
	} else {
		if (vfs_populate(ve) == 0) {
			vfs_population_foreach_reverse(ve, cve) {
				if (vfs_recurse(cve))
					do_playq_song_add_head(cve);
			}
		}
	}
}

void
playq_song_add_head(struct vfs_ent *ve)
{
	do_playq_song_add_head(ve);
	gui_playq_notify_done();
}

/**
 * @brief Recursively add a VFS entity to the tail of the playlist.
 */
static void
do_playq_song_add_tail(struct vfs_ent *ve)
{
	struct playq_entry *ent;
	struct vfs_ent *cve;

	if (vfs_playable(ve)) {
		ent = playq_entry_alloc(ve->filename, ve->name);

		if (ent != NULL) {
			TAILQ_INSERT_TAIL(&playq_list, ent, next);
			playq_len++;
			gui_playq_notify_post_insertion(playq_len - 1);
			sem_post(playq_available);
		}
	} else {
		if (vfs_populate(ve) == 0) {
			vfs_population_foreach(ve, cve) {
				if (vfs_recurse(cve))
					do_playq_song_add_tail(cve);
			}
		}
	}
}

void
playq_song_add_tail(struct vfs_ent *ve)
{
	do_playq_song_add_tail(ve);
	gui_playq_notify_done();
}

/**
 * @brief Seek the current song by a certain amount of time.
 */
static void
playq_cursong_seek(double len)
{
	do_seek += len;
	do_pause = 0;
#ifdef SIGCONT
	pthread_kill(playq_runner, SIGCONT);
#endif /* SIGCONT */
}

void
playq_cursong_seek_forward(void)
{
	playq_cursong_seek(5);
}

void
playq_cursong_seek_backward(void)
{
	playq_cursong_seek(-5);
}

void
playq_cursong_skip(void)
{
	do_skip = 1;
	do_pause = 0;
#ifdef SIGCONT
	pthread_kill(playq_runner, SIGCONT);
#endif /* SIGCONT */
}

void
playq_cursong_pause(void)
{
	if (do_pause) {
		/* Unpause */
		do_pause = 0;
#ifdef SIGCONT
		pthread_kill(playq_runner, SIGCONT);
#endif /* SIGCONT */
	} else
		/* Pause */
		do_pause = 1;
}

void
playq_song_fast_remove(struct playq_entry *pe, int index)
{
	if (sem_trywait(playq_available) != 0)
		return;

	gui_playq_notify_pre_removal(index);
	TAILQ_REMOVE(&playq_list, pe, next);
	playq_entry_free(pe);
	playq_len--;
	gui_playq_notify_done();
}

/**
 * @brief Recursively add a VFS entity before the specified entry.
 */
static void
do_playq_song_add_before(struct vfs_ent *ve, struct playq_entry *pe,
    int *idx_last)
{
	struct playq_entry *ent;
	struct vfs_ent *cve;

	if (vfs_playable(ve)) {
		ent = playq_entry_alloc(ve->filename, ve->name);

		if (ent != NULL) {
			TAILQ_INSERT_BEFORE(pe, ent, next);
			playq_len++;
			gui_playq_notify_post_insertion((*idx_last)++);
			sem_post(playq_available);
		}
	} else {
		if (vfs_populate(ve) == 0) {
			vfs_population_foreach(ve, cve) {
				if (vfs_recurse(cve))
					do_playq_song_add_before(cve, pe,
					    idx_last);
			}
		}
	}
}

void
playq_song_fast_add_before(struct vfs_ent *ve, struct playq_entry *pe,
    int index)
{
	do_playq_song_add_before(ve, pe, &index);
	gui_playq_notify_done();
}

/**
 * @brief Recursively add a VFS entity after the specified entry.
 */
static void
do_playq_song_add_after(struct vfs_ent *ve, struct playq_entry *pe,
    int index)
{
	struct playq_entry *ent;
	struct vfs_ent *cve;

	if (vfs_playable(ve)) {
		ent = playq_entry_alloc(ve->filename, ve->name);

		if (ent != NULL) {
			TAILQ_INSERT_AFTER(&playq_list, pe, ent, next);
			playq_len++;
			gui_playq_notify_post_insertion(index + 1);
			sem_post(playq_available);
		}
	} else {
		if (vfs_populate(ve) == 0) {
			vfs_population_foreach_reverse(ve, cve) {
				if (vfs_recurse(cve))
					do_playq_song_add_after(cve, pe,
					    index);
			}
		}
	}
}

void
playq_song_fast_add_after(struct vfs_ent *ve, struct playq_entry *pe,
    int index)
{
	do_playq_song_add_after(ve, pe, index);
	gui_playq_notify_done();
}

void
playq_song_randomize(void)
{
	int remaining, idx;
	struct playq_entry *pe;

	PLAYQ_LOCK;
	/*
	 * This routine randomizes the playlist. It works quite simple.
	 * It just constantly picks random entries from the beginning
	 * and appends them to the end of the playlist.
	 */
	for (remaining = playq_len; remaining > 0; remaining--) {
		/* Fetch a random object */
		pe = TAILQ_FIRST(&playq_list);
		for (idx = random() % remaining; idx > 0; idx--) {
			pe = TAILQ_NEXT(pe, next);
			assert(pe != NULL);
		}

		/* Remove it and place it at the end */
		TAILQ_REMOVE(&playq_list, pe, next);
		TAILQ_INSERT_TAIL(&playq_list, pe, next);
	}

	gui_playq_notify_post_randomization();
	gui_playq_notify_done();
	PLAYQ_UNLOCK;
}
