/*
 * Copyright (c) 2006 Martijn van den Dungen <martijn@il.fontys.nl>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: vfs.c 459 2006-07-28 00:08:22Z ed $
 */
/**
 * @file vfs/vfs.c
 */

#include "config.h"

#ifdef HAVE_LIBGEN_H
#include <libgen.h>
#endif /* HAVE_LIBGEN_H */
#ifdef HAVE_WINDOWS_H
#include <windows.h>
#endif /* HAVE_WINDOWS_H */
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#ifdef HAVE_PWD_H
#include <pwd.h>
#endif /* HAVE_PWD_H */
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <trans.h>
#include <configfile/configfile.h>
#include <vfs/vfs.h>
#include <vfs/modules/vfs_file.h>
#include <vfs/modules/vfs_dir.h>
#include <vfs/modules/vfs_m3u.h>
#ifdef _WIN32
#include <vfs/modules/vfs_win32.h>
#endif /* _WIN32 */

/**
 * @brief List of available virtual filesystem modules and their options
 *        and functions.
 */
static struct vfs_module modules[] = {
#ifdef _WIN32
	/*
	 * Leave this module at the top of the list. This entry fakes
	 * our 'root' directory on Windows. The second module is to
	 * implement Windows root directories of harddisks, as D: is not
	 * the same as D:\. We do not want to stat them, otherwise
	 * Windows will search for floppies.
	 */
	{ vfs_win32_root_open, NULL, vfs_win32_root_populate, 0, 0, VFS_SORT_FIRST, PATH_SEP },
	{ vfs_win32_disk_open, NULL, vfs_win32_disk_populate, 0, 0, VFS_SORT_FIRST, PATH_SEP },
#endif /* !_WIN32 */
	{ vfs_m3u_open, NULL, vfs_m3u_populate, 0, 0, VFS_SORT_LAST, '@' },
	/*
	 * Leave these two rules at the bottom of the list. They have
	 * the weakest matching rules.
	 */
	{ vfs_dir_open, NULL, vfs_dir_populate, 1, 0, VFS_SORT_FIRST, PATH_SEP },
	{ vfs_file_open, NULL, NULL, 1, 1, VFS_SORT_LAST, '\0' },
};
/**
 * @brief The number of virtual file system modules currently available
 *        in the application.
 */
#define NUM_MODULES (sizeof(modules) / sizeof(struct vfs_module))

int
vfs_lockup(void)
{
#ifdef HAVE_CHROOT
	const char *root;
#endif /* HAVE_CHROOT */
#ifdef HAVE_GETPWNAM
	const char *user;
	struct passwd *pw = NULL;

	user = config_getopt("vfs.lockup.user");
	if (user[0] != '\0') {
		pw = getpwnam(user);
		if (pw == NULL) {
			fprintf(stderr, _("Unknown user: %s\n"), user);
			return (-1);
		}
	}
#endif /* HAVE_GETPWNAM */

#ifdef HAVE_CHROOT
	root = config_getopt("vfs.lockup.chroot");

	if (root[0] != '\0') {
		/* Try to lock ourselves in */
		if (chroot(root) != 0) {
			fprintf(stderr, _("Unable to chroot in %s\n"), root);
			return (-1);
		}

		chdir("/");
	}
#endif /* HAVE_CHROOT */

#ifdef HAVE_GETPWNAM
	if (pw != NULL) {
		if (setgid(pw->pw_gid) != 0) {
			fprintf(stderr, _("Unable to change to group %d\n"),
			    pw->pw_gid);
			return (-1);
		}
		if (setuid(pw->pw_uid) != 0) {
			fprintf(stderr, _("Unable to change to user %d\n"),
			    pw->pw_uid);
			return (-1);
		}
	}
#endif /* HAVE_GETPWNAM */

	return (0);
}

/**
 * @brief Deallocates the data structures for a VFS entity
 */
static void
vfs_dealloc(struct vfs_ent *ve)
{
	free(ve->filename);
	free(ve);
}

struct vfs_ent *
vfs_open(const char *filename)
{
	struct vfs_ent *ve;
	struct stat fs;
	int i;
	char rpath[PATH_MAX];

#ifdef _WIN32
	/*
	 * XXX: Trailing PATH_SEP's
	 */
	if (filename[PATH_MIN] == '\0' ||
	    strcmp(filename, WIN32_ROOT) == 0) {
		strcpy(rpath, filename);
		goto checkdone;
	}
#endif /* _WIN32 */

#if defined(HAVE_REALPATH)
	realpath(filename, rpath);
#elif defined(HAVE_WINDOWS_H)
	LPSTR lpf;
	GetFullPathNameA(filename, PATH_MAX, rpath, &lpf);
#else
#error Cannot determine real pathname
#endif

#ifdef HAVE_LSTAT
	if (lstat(rpath, &fs) != 0)
#else /* !HAVE_LSTAT */
	if (stat(rpath, &fs) != 0)
#endif /* HAVE_STAT */
		return (NULL);

	/* Make sure we don't run into fifo's and such */
	if (!S_ISREG(fs.st_mode) && !S_ISDIR(fs.st_mode))
		return (NULL);
#ifdef _WIN32
checkdone:
#endif /* _WIN32 */
	if ((ve = malloc(sizeof(struct vfs_ent))) == NULL)
		return (NULL);
	
	/*
	 * Fill our structure with the basic stuff
	 */
	memset(ve, 0x00000000, sizeof(struct vfs_ent));

	ve->filename = strdup(rpath);
	ve->name = strrchr(ve->filename, PATH_SEP);
	if ((ve->name == NULL) || (ve->name[1] == '\0'))
		ve->name = ve->filename;
	else
		ve->name++;

	ve->refcount = 1;

	TAILQ_INIT(&ve->population);

	for (i = 0; i < NUM_MODULES; i++) {
		if (modules[i].v_open(ve, S_ISDIR(fs.st_mode)) == 0) {
			ve->vmod = &modules[i];
			break;
		}
	}

	if (ve->vmod == NULL) {
		vfs_dealloc(ve);
		return (NULL);
	}

	return (ve);
}

struct vfs_ent *
vfs_parent(struct vfs_ent *ve)
{
	char *fn;
	int len;
	struct vfs_ent *ret;
	
#ifdef _WIN32
	/*  There's nothing below the root directory */
	if (strcmp(ve->filename, WIN32_ROOT) == 0)
		return (NULL);

	/* Win32: Evil hack to strand down at the root */
	if (ve->filename[PATH_MIN] == '\0')
		return vfs_open(WIN32_ROOT);
#else /* !_WIN32 */
	/*  There's nothing below the root directory */
	if (ve->filename[1] == '\0')
		return (NULL);
#endif /* _WIN32 */

	fn = strdup(ve->filename);
	
	/* Get the parent filename */
	for (len = strlen(fn);
	    (len > PATH_MIN) && (fn[len] != PATH_SEP); len--);
	fn[len] = '\0';

	ret = vfs_open(fn);
	free(fn);

	return (ret);
}

struct vfs_ent *
vfs_dup(struct vfs_ent *ve)
{
	ve->refcount++;

	return (ve);
}

void
vfs_close(struct vfs_ent *ve)
{
	struct vfs_ent *cur;

	if (--ve->refcount > 0)
		return;

	while (!TAILQ_EMPTY(&ve->population)) {
		cur = TAILQ_FIRST(&ve->population);
		TAILQ_REMOVE(&ve->population, cur, next);
		vfs_close(cur);
	}

	if (ve->vmod->v_close != NULL)
		ve->vmod->v_close(ve);
	
	vfs_dealloc(ve);
}

int
vfs_populate(struct vfs_ent *ve)
{
	/* Some object cannot be populated */
	if (ve->vmod->v_populate == NULL)
		return (-1);

	/* Don't fetch double data */
	if (!TAILQ_EMPTY(&ve->population))
		return (0);

	return ve->vmod->v_populate(ve);
}
